@@ -12,10 +12,10 @@ from pathlib import Path
 from urllib.parse import urljoin
 
 import cssselect2
-import html5lib
 import tinycss2
+import tinyhtml5
 
-VERSION = __version__ = '62.1'
+VERSION = __version__ = '63.1'
 
 #: Default values for command-line and Python API options. See
 #: :func:`__main__.main` to learn more about specific options for
@@ -45,6 +45,9 @@ VERSION = __version__ = '62.1'
 #:     Whether custom HTML metadata should be stored in the generated PDF.
 #: :param bool presentational_hints:
 #:     Whether HTML presentational hints are followed.
+#: :param bool srgb:
+#:     Whether sRGB color profile should be included and set as default for
+#:     device-dependant RGB colors.
 #: :param bool optimize_images:
 #:     Whether size of embedded images should be optimized, with no quality
 #:     loss.
@@ -71,6 +74,7 @@ DEFAULT_OPTIONS = {
     'uncompressed_pdf': False,
     'custom_metadata': False,
     'presentational_hints': False,
+    'srgb': False,
     'optimize_images': False,
     'jpeg_quality': None,
     'dpi': None,
@@ -80,8 +84,8 @@ DEFAULT_OPTIONS = {
 }
 
 __all__ = [
-    'HTML', 'CSS', 'DEFAULT_OPTIONS', 'Attachment', 'Document', 'Page',
-    'default_url_fetcher', 'VERSION', '__version__']
+    'CSS', 'DEFAULT_OPTIONS', 'HTML', 'VERSION', 'Attachment', 'Document', 'Page',
+    '__version__', 'default_url_fetcher']
 
 
 # Import after setting the version, as the version is used in other modules
@@ -107,7 +111,7 @@ def _find_base_url(html_document, fallback_base_url):
 
 
 class HTML:
-    """HTML document parsed by html5lib.
+    """HTML document parsed by tinyhtml5.
 
     You can just create an instance with a positional argument:
     ``doc = HTML(something)``
@@ -165,14 +169,14 @@ class HTML:
             guess, filename, url, file_obj, string, base_url, url_fetcher)
         with result as (source_type, source, base_url, protocol_encoding):
             if isinstance(source, str):
-                result = html5lib.parse(source, namespaceHTMLElements=False)
+                result = tinyhtml5.parse(source, namespace_html_elements=False)
             else:
-                kwargs = {'namespaceHTMLElements': False}
+                kwargs = {'namespace_html_elements': False}
                 if protocol_encoding is not None:
                     kwargs['transport_encoding'] = protocol_encoding
                 if encoding is not None:
                     kwargs['override_encoding'] = encoding
-                result = html5lib.parse(source, **kwargs)
+                result = tinyhtml5.parse(source, **kwargs)
         self.base_url = _find_base_url(result, base_url)
         self.url_fetcher = url_fetcher
         self.media_type = media_type
@@ -210,6 +214,8 @@ class HTML:
         :returns: A :class:`document.Document` object.
 
         """
+        for unknown in set(options) - set(DEFAULT_OPTIONS):
+            LOGGER.warning('Unknown rendering option: %s.', unknown)
         new_options = DEFAULT_OPTIONS.copy()
         new_options.update(options)
         options = new_options
