@@ -3,10 +3,10 @@
 from math import pi
 from urllib.parse import unquote
 
-from tinycss2.color3 import parse_color
+from tinycss2.color4 import parse_color
 
 from ..logger import LOGGER
-from ..text.ffi import ffi, pango, units_to_double
+from ..text.ffi import FROM_UNITS, ffi, pango
 from ..text.line_break import Layout, first_line_metrics
 from ..urls import get_link_attribute
 from .properties import INITIAL_VALUES, ZERO_PIXELS, Dimension
@@ -362,6 +362,7 @@ def border_width(style, name, value):
 
 
 @register_computer('border-image-slice')
+@register_computer('mask-border-slice')
 def border_image_slice(style, name, values):
     """Compute the ``border-image-slice`` property."""
     computed_values = []
@@ -385,6 +386,7 @@ def border_image_slice(style, name, values):
 
 
 @register_computer('border-image-width')
+@register_computer('mask-border-width')
 def border_image_width(style, name, values):
     """Compute the ``border-image-width`` property."""
     computed_values = []
@@ -404,6 +406,7 @@ def border_image_width(style, name, values):
 
 
 @register_computer('border-image-outset')
+@register_computer('mask-border-outset')
 def border_image_outset(style, name, values):
     """Compute the ``border-image-outset`` property."""
     computed_values = [
@@ -419,6 +422,7 @@ def border_image_outset(style, name, values):
 
 
 @register_computer('border-image-repeat')
+@register_computer('mask-border-repeat')
 def border_image_repeat(style, name, values):
     """Compute the ``border-image-repeat`` property."""
     return (values * 2) if len(values) == 1 else values
@@ -445,7 +449,7 @@ def gap(style, name, value):
     """Compute the ``*-gap`` properties."""
     if value == 'normal':
         return value
-    return length(style, name, value, pixels_only=True)
+    return length(style, name, value)
 
 
 def _content_list(style, values):
@@ -811,9 +815,9 @@ def character_ratio(style, character):
     logical_extents = ffi.new('PangoRectangle *')
     pango.pango_layout_line_get_extents(line, ink_extents, logical_extents)
     if character == 'x':
-        measure = -units_to_double(ink_extents.y)
+        measure = -ink_extents.y * FROM_UNITS
     else:
-        measure = units_to_double(logical_extents.width)
+        measure = logical_extents.width * FROM_UNITS
     ffi.release(ink_extents)
     ffi.release(logical_extents)
 
