@@ -3,7 +3,7 @@
 import functools
 
 from tinycss2.ast import DimensionToken, IdentToken, NumberToken
-from tinycss2.color3 import parse_color
+from tinycss2.color4 import parse_color
 
 from ..properties import INITIAL_VALUES
 from .descriptors import expand_font_variant
@@ -18,8 +18,8 @@ from .properties import ( # isort:skip
     border_width, box, column_count, column_width, flex_basis, flex_direction,
     flex_grow_shrink, flex_wrap, font_family, font_size, font_stretch,
     font_style, font_weight, gap, grid_line, grid_template, line_height,
-    list_style_image, list_style_position, list_style_type, other_colors,
-    overflow_wrap, validate_non_shorthand)
+    list_style_image, list_style_position, list_style_type, mask_border_mode,
+    other_colors, overflow_wrap, validate_non_shorthand)
 
 EXPANDERS = {}
 
@@ -350,6 +350,72 @@ def expand_border_image(tokens, name, base_url):
             raise InvalidValues
 
 
+@expander('mask-border')
+@generic_expander('-outset', '-repeat', '-slice', '-source', '-width', '-mode',
+                  wants_base_url=True)
+def expand_mask_border(tokens, name, base_url):
+    """Expand the ``mask-border-*`` shorthand properties.
+
+    See https://drafts.fxtf.org/css-masking/#the-mask-border
+
+    """
+    tokens = list(tokens)
+    while tokens:
+        if border_image_source(tokens[:1], base_url):
+            yield '-source', [tokens.pop(0)]
+        elif mask_border_mode(tokens[:1]):
+            yield '-mode', [tokens.pop(0)]
+        elif border_image_repeat(tokens[:1]):
+            repeats = [tokens.pop(0)]
+            while tokens and border_image_repeat(tokens[:1]):
+                repeats.append(tokens.pop(0))
+            yield '-repeat', repeats
+        elif border_image_slice(tokens[:1]) or get_keyword(tokens[0]) == 'fill':
+            slices = [tokens.pop(0)]
+            while tokens and border_image_slice(slices + tokens[:1]):
+                slices.append(tokens.pop(0))
+            yield '-slice', slices
+            if tokens and tokens[0].type == 'literal' and tokens[0].value == '/':
+                # slices / *
+                tokens.pop(0)
+            else:
+                # slices other
+                continue
+            if not tokens:
+                # slices /
+                raise InvalidValues
+            if border_image_width(tokens[:1]):
+                widths = [tokens.pop(0)]
+                while tokens and border_image_width(widths + tokens[:1]):
+                    widths.append(tokens.pop(0))
+                yield '-width', widths
+                if tokens and tokens[0].type == 'literal' and tokens[0].value == '/':
+                    # slices / widths / slash *
+                    tokens.pop(0)
+                else:
+                    # slices / widths other
+                    continue
+            elif tokens and tokens[0].type == 'literal' and tokens[0].value == '/':
+                # slices / / *
+                tokens.pop(0)
+            else:
+                # slices / other
+                raise InvalidValues
+            if not tokens:
+                # slices / * /
+                raise InvalidValues
+            if border_image_outset(tokens[:1]):
+                outsets = [tokens.pop(0)]
+                while tokens and border_image_outset(outsets + tokens[:1]):
+                    outsets.append(tokens.pop(0))
+                yield '-outset', outsets
+            else:
+                # slash / * / other
+                raise InvalidValues
+        else:
+            raise InvalidValues
+
+
 @expander('background')
 def expand_background(tokens, name, base_url):
     """Expand the ``background`` shorthand property.
@@ -830,17 +896,22 @@ def expand_grid(tokens, name):
     templates = {'row': [], 'column': []}
     iterable = zip(split_tokens, templates.items())
     for tokens, (track, track_templates) in iterable:
+        auto_flow_token = False
         for token in tokens:
             if get_keyword(token) == 'dense':
                 if dense or (auto_track and auto_track != track):
                     raise InvalidValues
                 dense = token
-            elif get_keyword(token) == 'auto-flow':
-                if auto_track:
-                    raise InvalidValues
                 auto_track = track
-            else:
+            elif get_keyword(token) == 'auto-flow':
+                if auto_flow_token or (auto_track and auto_track != track):
+                    raise InvalidValues
+                auto_flow_token = True
+                auto_track = track
+            elif token == tokens[-1]:
                 track_templates.append(token)
+            else:
+                raise InvalidValues
     if not auto_track:
         raise InvalidValues
     non_auto_track = 'row' if auto_track == 'column' else 'column'
@@ -900,7 +971,7 @@ def expand_grid_column_row(tokens, name):
 def expand_grid_area(tokens, name):
     """Expand the ``grid-area`` property."""
     tokens_list = _expand_grid_column_row_area(tokens, 4)
-    sides = ('row-start', 'row-end', 'column-start', 'column-end')
+    sides = ('row-start', 'column-start', 'row-end', 'column-end')
     for tokens, side in zip(tokens_list, sides):
         yield f'grid-{side}', tokens
 
