@@ -7,7 +7,7 @@ See https://www.w3.org/TR/CSS21/propidx.html and various CSS3 modules.
 from math import inf
 
 from tinycss2 import parse_component_value_list
-from tinycss2.color3 import parse_color
+from tinycss2.color4 import parse_color
 
 from .. import computed_values
 from ..properties import KNOWN_PROPERTIES, ZERO_PIXELS, Dimension
@@ -135,7 +135,7 @@ def other_colors(token):
 @single_token
 def outline_color(token):
     if get_keyword(token) == 'invert':
-        return 'currentColor'
+        return 'currentcolor'
     else:
         return parse_color(token)
 
@@ -158,7 +158,7 @@ def empty_cells(keyword):
 def color(token):
     """``*-color`` and ``color`` properties validation."""
     result = parse_color(token)
-    if result == 'currentColor':
+    if result == 'currentcolor':
         return 'inherit'
     else:
         return result
@@ -421,7 +421,8 @@ def border_width(token):
         return keyword
 
 
-@property(wants_base_url=True)
+@property('border-image-source', wants_base_url=True)
+@property('mask-border-source', wants_base_url=True)
 @single_token
 def border_image_source(token, base_url):
     if get_keyword(token) == 'none':
@@ -429,7 +430,8 @@ def border_image_source(token, base_url):
     return get_image(token, base_url)
 
 
-@property()
+@property('border-image-slice')
+@property('mask-border-slice')
 def border_image_slice(tokens):
     values = []
     fill = False
@@ -449,7 +451,8 @@ def border_image_slice(tokens):
         return tuple(values)
 
 
-@property()
+@property('border-image-width')
+@property('mask-border-width')
 def border_image_width(tokens):
     values = []
     for token in tokens:
@@ -467,7 +470,8 @@ def border_image_width(tokens):
         return tuple(values)
 
 
-@property()
+@property('border-image-outset')
+@property('mask-border-outset')
 def border_image_outset(tokens):
     values = []
     for token in tokens:
@@ -483,7 +487,8 @@ def border_image_outset(tokens):
         return tuple(values)
 
 
-@property()
+@property('border-image-repeat')
+@property('mask-border-repeat')
 def border_image_repeat(tokens):
     if 1 <= len(tokens) <= 2:
         keywords = tuple(get_keyword(token) for token in tokens)
@@ -491,6 +496,12 @@ def border_image_repeat(tokens):
             return keywords
 
 
+@property()
+@single_keyword
+def mask_border_mode(keyword):
+    return keyword in ('luminance', 'alpha')
+
+
 @property(unstable=True)
 @single_token
 def column_width(token):
@@ -664,7 +675,7 @@ def width_height(token):
 @single_token
 def gap(token):
     """Validation for the ``column-gap`` and ``row-gap`` properties."""
-    length = get_length(token, negative=False)
+    length = get_length(token, percentage=True, negative=False)
     if length:
         return length
     keyword = get_keyword(token)
