@@ -14,6 +14,7 @@ from xml.etree import ElementTree
 
 import pydyf
 from PIL import Image, ImageFile, ImageOps
+from tinycss2.color4 import parse_color
 
 from . import DEFAULT_OPTIONS
 from .layout.percent import percentage
@@ -67,6 +68,12 @@ class RasterImage:
         self.ratio = (self.width / self.height) if self.height != 0 else inf
         self.optimize = optimize = options['optimize_images']
 
+        # The presence of the APP14 segment indicates an Adobe image with
+        # inverted CMYK data. Specify a Decode Array to invert it again back to
+        # normal. See https://github.com/Kozea/WeasyPrint/pull/2179.
+        app14 = getattr(original_pillow_image, 'app', {}).get('APP14')
+        self.invert_colors = self.mode == 'CMYK' and app14 is not None
+
         if pillow_image.format in ('JPEG', 'MPO'):
             self.format = 'JPEG'
             if image_data is None or optimize or jpeg_quality is not None:
@@ -108,11 +115,11 @@ class RasterImage:
             concrete_width, 0, 0, -concrete_height, 0, concrete_height)
         stream.draw_x_object(image_name)
 
-    def cache_image_data(self, data, filename=None, alpha=False):
+    def cache_image_data(self, data, filename=None, slot='source'):
         if filename:
             return LazyLocalImage(filename)
         else:
-            key = f'{self.id}{int(alpha)}{self._dpi or ""}'
+            key = f'{self.id}-{slot}-{self._dpi or ""}'
             return LazyImage(self._cache, key, data)
 
     def get_x_object(self, interpolate, dpi_ratio):
@@ -150,6 +157,8 @@ class RasterImage:
         })
 
         if self.format == 'JPEG':
+            if self.invert_colors:
+                extra['Decode'] = pydyf.Array((1, 0) * 4)
             extra['Filter'] = '/DCTDecode'
             return pydyf.Stream([self.image_data], extra)
 
@@ -175,7 +184,7 @@ class RasterImage:
             png_data = self._get_png_data(pillow_image)
             # Save alpha channel as mask
             alpha_data = self._get_png_data(alpha)
-            stream = self.cache_image_data(alpha_data, alpha=True)
+            stream = self.cache_image_data(alpha_data, slot='streamalpha')
             extra['SMask'] = pydyf.Stream([stream], extra={
                 'Filter': '/FlateDecode',
                 'Type': '/XObject',
@@ -194,7 +203,7 @@ class RasterImage:
             png_data = self._get_png_data(
                 Image.open(io.BytesIO(self.image_data.data)))
 
-        return pydyf.Stream([self.cache_image_data(png_data)], extra)
+        return pydyf.Stream([self.cache_image_data(png_data, slot='stream')], extra)
 
     @staticmethod
     def _get_png_data(pillow_image):
@@ -277,8 +286,9 @@ class SVGImage:
             self._svg.draw(
                 stream, concrete_width, concrete_height, self._base_url,
                 self._url_fetcher, self._context)
-        except BaseException:
+        except BaseException as exception:
             LOGGER.error('Failed to render SVG image %s', self._base_url)
+            LOGGER.debug('Error while rendering SVG image:', exc_info=exception)
 
 
 def get_image_from_uri(cache, url_fetcher, options, url, forced_mime_type=None,
@@ -326,13 +336,14 @@ def get_image_from_uri(cache, url_fetcher, options, url, forced_mime_type=None,
                     raise ImageLoadingError.from_exception(raster_exception)
             else:
                 # Store image id to enable cache in Stream.add_image
-                image_id = md5(url.encode()).hexdigest()
+                image_id = md5(url.encode(), usedforsecurity=False).hexdigest()
                 image = RasterImage(
                     pillow_image, image_id, string, filename, cache,
                     orientation, options)
 
     except (URLFetchingError, ImageLoadingError) as exception:
         LOGGER.error('Failed to load image at %r: %s', url, exception)
+        LOGGER.debug('Error while loading image:', exc_info=exception)
         image = None
 
     cache[url] = image
@@ -431,6 +442,7 @@ def gradient_average_color(colors, positions):
     """
     https://drafts.csswg.org/css-images-3/#gradient-average-color
     """
+    # TODO: handle color spaces.
     nb_stops = len(positions)
     assert nb_stops > 1
     assert nb_stops == len(colors)
@@ -451,9 +463,13 @@ def gradient_average_color(colors, positions):
             result_g += premul_g[j] * weight
             result_b += premul_b[j] * weight
             result_a += alpha[j] * weight
-    # Un-premultiply:
-    return (result_r / result_a, result_g / result_a,
-            result_b / result_a, result_a) if result_a != 0 else (0, 0, 0, 0)
+    # Un-premultiply.
+    if result_a == 0:
+        return parse_color('transparent')
+    else:
+        return parse_color(
+            f'rgb({result_r / result_a * 255} {result_g / result_a * 255} '
+            f'{result_b / result_a * 255}/{ result_a })')
 
 
 class Gradient:
@@ -476,10 +492,7 @@ class Gradient:
 
         if type_ == 'solid':
             stream.rectangle(0, 0, concrete_width, concrete_height)
-            red, green, blue, alpha = colors[0]
-            stream.set_color_rgb(red, green, blue)
-            if alpha != 1:
-                stream.set_alpha(alpha, stroke=False)
+            stream.set_color(colors[0])
             stream.fill()
             return
 
@@ -531,7 +544,7 @@ class Gradient:
             alpha_stream.transform(d=scale_y)
             alpha_stream.stream = [f'/{alpha_shading.id} sh']
 
-        stream.shading(shading.id)
+        stream.paint_shading(shading.id)
 
     def layout(self, width, height):
         """Get layout information about the gradient.
@@ -619,9 +632,9 @@ class LinearGradient(Gradient):
                 for i in range(len(positions) - 1)]
 
             # Create cycles used to add colors
-            next_steps = cycle([0] + position_steps)
+            next_steps = cycle((0, *position_steps))
             next_colors = cycle(colors)
-            previous_steps = cycle([0] + position_steps[::-1])
+            previous_steps = cycle((0, *position_steps[::-1]))
             previous_colors = cycle(colors[::-1])
 
             # Add colors after last step
@@ -820,8 +833,7 @@ class RadialGradient(Gradient):
                 new_positions = [
                     position - 1 - full_repeat for position
                     in original_positions[-(i - 1):]]
-                positions = (
-                    [ratio - 1 - full_repeat] + new_positions + positions)
+                positions = (ratio - 1 - full_repeat, *new_positions, *positions)
                 return points, positions, colors
 
     def _resolve_size(self, width, height, center_x, center_y):
