@@ -1,9 +1,10 @@
 """Layout for pages and CSS3 margin boxes."""
 
 import copy
+from collections import namedtuple
 from math import inf
 
-from ..css import PageType, computed_from_cascaded
+from ..css import computed_from_cascaded
 from ..formatting_structure import boxes, build
 from ..logger import PROGRESS_LOGGER
 from .absolute import absolute_box_layout, absolute_layout
@@ -13,6 +14,8 @@ from .min_max import handle_min_max_height, handle_min_max_width
 from .percent import resolve_percentages
 from .preferred import max_content_width, min_content_width
 
+PageType = namedtuple('PageType', ['side', 'blank', 'name', 'index', 'group_index'])
+
 
 class OrientedBox:
     @property
@@ -604,7 +607,8 @@ def make_page(context, root_box, page_type, resume_at, page_number,
 
     # TODO: handle cases where the root element is something else.
     # See https://www.w3.org/TR/CSS21/visuren.html#dis-pos-flo
-    assert isinstance(root_box, (boxes.BlockBox, boxes.FlexContainerBox))
+    assert isinstance(root_box, (
+        boxes.BlockBox, boxes.FlexContainerBox, boxes.GridContainerBox))
     context.create_block_formatting_context()
     context.current_page = page_number
     context.current_page_footnotes = []
@@ -796,7 +800,7 @@ def set_page_type_computed_styles(page_type, html, style_for):
                 base_url=html.base_url)
 
 
-def remake_page(index, context, root_box, html):
+def remake_page(index, group_index, context, root_box, html):
     """Return one laid out page without margin boxes.
 
     Start with the initial values from ``context.page_maker[index]``.
@@ -816,7 +820,6 @@ def remake_page(index, context, root_box, html):
     # Don't modify actual page_maker[index] values!
     # TODO: should we store (and reuse) page_type in the page_maker?
     page_state = copy.deepcopy(initial_page_state)
-    first = index == 0
     if initial_next_page['break'] in ('left', 'right'):
         next_page_side = initial_next_page['break']
     elif initial_next_page['break'] in ('recto', 'verso'):
@@ -829,10 +832,14 @@ def remake_page(index, context, root_box, html):
         (next_page_side == 'left' and right_page) or
         (next_page_side == 'right' and not right_page) or
         (context.reported_footnotes and initial_resume_at is None))
-    next_page_name = '' if blank else initial_next_page['page']
+    name = '' if blank else initial_next_page['page']
+    if name not in group_index:
+        group_index.clear()
+        group_index[name] = 0
     side = 'right' if right_page else 'left'
-    page_type = PageType(side, blank, first, index, name=next_page_name)
+    page_type = PageType(side, blank, name, index, group_index[name])
     set_page_type_computed_styles(page_type, html, context.style_for)
+    group_index[name] += 1
 
     context.forced_break = (
         initial_next_page['break'] != 'any' or initial_next_page['page'])
@@ -891,6 +898,7 @@ def make_all_pages(context, root_box, html, pages):
     """
     i = 0
     reported_footnotes = None
+    group_i = {'': 0}
     while True:
         remake_state = context.page_maker[i][-1]
         if (len(pages) == 0 or
@@ -902,7 +910,7 @@ def make_all_pages(context, root_box, html, pages):
             remake_state['pages_wanted'] = False
             remake_state['anchors'] = []
             remake_state['content_lookups'] = []
-            page, resume_at = remake_page(i, context, root_box, html)
+            page, resume_at = remake_page(i, group_i, context, root_box, html)
             reported_footnotes = context.reported_footnotes
             yield page
         else:
