@@ -6,9 +6,7 @@ from math import inf
 import pyphen
 
 from .constants import LST_TO_ISO, PANGO_WRAP_MODE
-from .ffi import (
-    ffi, gobject, pango, pangoft2, unicode_to_char_p, units_from_double,
-    units_to_double)
+from .ffi import FROM_UNITS, TO_UNITS, ffi, gobject, pango, pangoft2, unicode_to_char_p
 from .fonts import font_features, get_font_description
 
 
@@ -20,8 +18,8 @@ def line_size(line, style):
     """
     logical_extents = ffi.new('PangoRectangle *')
     pango.pango_layout_line_get_extents(line, ffi.NULL, logical_extents)
-    width = units_to_double(logical_extents.width)
-    height = units_to_double(logical_extents.height)
+    width = logical_extents.width * FROM_UNITS
+    height = logical_extents.height * FROM_UNITS
     ffi.release(logical_extents)
     if style['letter_spacing'] != 'normal':
         width += style['letter_spacing']
@@ -51,7 +49,7 @@ def first_line_metrics(first_line, text, layout, resume_at, space_collapse,
         length = first_line.length if first_line is not None else 0
 
     width, height = line_size(first_line, style)
-    baseline = units_to_double(pango.pango_layout_get_baseline(layout.layout))
+    baseline = pango.pango_layout_get_baseline(layout.layout) * FROM_UNITS
     layout.deactivate()
     return layout, length, resume_at, width, height, baseline
 
@@ -106,15 +104,15 @@ class Layout:
                 pango.pango_context_get_metrics(
                     pango_context, font_description, self.language),
                 pango.pango_font_metrics_unref)
-            self.ascent = units_to_double(
+            self.ascent = FROM_UNITS * (
                 pango.pango_font_metrics_get_ascent(metrics))
-            self.underline_position = units_to_double(
+            self.underline_position = FROM_UNITS * (
                 pango.pango_font_metrics_get_underline_position(metrics))
-            self.strikethrough_position = units_to_double(
+            self.strikethrough_position = FROM_UNITS * (
                 pango.pango_font_metrics_get_strikethrough_position(metrics))
-            self.underline_thickness = units_to_double(
+            self.underline_thickness = FROM_UNITS * (
                 pango.pango_font_metrics_get_underline_thickness(metrics))
-            self.strikethrough_thickness = units_to_double(
+            self.strikethrough_thickness = FROM_UNITS * (
                 pango.pango_font_metrics_get_strikethrough_thickness(metrics))
         else:
             self.ascent = None
@@ -178,7 +176,7 @@ class Layout:
                 pango.pango_attr_list_change(attr_list, attr)
 
             if letter_spacing:
-                letter_spacing = units_from_double(letter_spacing)
+                letter_spacing = int(letter_spacing * TO_UNITS)
                 add_attr(0, len(bytestring), letter_spacing)
 
             if word_spacing:
@@ -188,8 +186,7 @@ class Layout:
                     text, bytestring = unicode_to_char_p(self.text)
                     pango.pango_layout_set_text(self.layout, text, -1)
 
-                space_spacing = (
-                    units_from_double(word_spacing) + letter_spacing)
+                space_spacing = int(word_spacing * TO_UNITS + letter_spacing)
                 position = bytestring.find(b' ')
                 # Pango gives only half of word-spacing on boundaries
                 boundary_positions = (0, len(bytestring) - 1)
@@ -242,8 +239,7 @@ def create_layout(text, style, context, max_width, justification_spacing):
     # signed integer. Treat bigger values same as None: unconstrained width.
     text_wrap = style['white_space'] in ('normal', 'pre-wrap', 'pre-line')
     if max_width is not None and text_wrap and max_width < 2 ** 21:
-        pango.pango_layout_set_width(
-            layout.layout, units_from_double(max(0, max_width)))
+        pango.pango_layout_set_width(layout.layout, int(max(0, max_width) * TO_UNITS))
 
     layout.set_text(text)
     return layout
@@ -316,7 +312,6 @@ def split_first_line(text, style, context, max_width, justification_spacing,
     first_line_text = text.encode()[:resume_index].decode()
     first_line_fits = (
         first_line_width <= max_width or
-        ' ' in first_line_text.strip() or
         can_break_text(first_line_text.strip(), style['lang']))
     if first_line_fits:
         # The first line fits but may have been cut too early by Pango
@@ -326,9 +321,10 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         first_line_text = ''
         second_line_text = text
 
-    next_word = second_line_text.split(' ', 1)[0]
+    break_point = next_break_point(second_line_text, style['lang'])
+    next_word = second_line_text[:break_point].rstrip(' ')
     if next_word:
-        if space_collapse:
+        if space_collapse and second_line_text[break_point or -1] == ' ':
             # next_word might fit without a space afterwards
             # only try when space collapsing is allowed
             new_first_line_text = first_line_text + next_word
@@ -361,7 +357,7 @@ def split_first_line(text, style, context, max_width, justification_spacing,
 
     auto_hyphenation = manual_hyphenation = False
     if hyphens != 'none':
-        manual_hyphenation = soft_hyphen in first_line_text + next_word
+        manual_hyphenation = soft_hyphen in first_line_text + second_line_text
     if hyphens == 'auto' and lang:
         next_word_boundaries = get_next_word_boundaries(second_line_text, lang)
         if next_word_boundaries:
@@ -391,19 +387,11 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         # hyphen and add the missing hyphen
         if first_line_text.endswith(soft_hyphen):
             # The first line has been split on a soft hyphen
-            if ' ' in first_line_text:
-                first_line_text, next_word = first_line_text.rsplit(' ', 1)
-                next_word = f' {next_word}'
-                layout.set_text(first_line_text)
-                first_line, _ = layout.get_first_line()
-                resume_index = len((f'{first_line_text} ').encode())
-            else:
-                first_line_text, next_word = '', first_line_text
+            first_line_text, second_line_text = '', first_line_text
         soft_hyphen_indexes = [
-            match.start() for match in re.finditer(soft_hyphen, next_word)]
+            match.start() for match in re.finditer(soft_hyphen, second_line_text)]
         soft_hyphen_indexes.reverse()
-        dictionary_iterations = [
-            next_word[:i + 1] for i in soft_hyphen_indexes]
+        dictionary_iterations = [second_line_text[:i+1] for i in soft_hyphen_indexes]
         start_word = 0
     elif auto_hyphenation:
         dictionary_key = (lang, left, right, total)
@@ -430,19 +418,12 @@ def split_first_line(text, style, context, max_width, justification_spacing,
             new_first_line, index = new_layout.get_first_line()
             new_first_line_width, _ = line_size(new_first_line, style)
             new_space = max_width - new_first_line_width
-            hyphenated = (
-                index is None and (
-                    new_space >= 0 or
-                    first_word_part == dictionary_iterations[-1]))
+            hyphenated = index is None and (
+                new_space >= 0 or first_word_part == dictionary_iterations[-1])
             if hyphenated:
                 layout = new_layout
                 first_line = new_first_line
                 resume_index = len(new_first_line_text.encode())
-                if text[len(new_first_line_text)] == soft_hyphen:
-                    # Recreate the layout with no max_width to be sure that
-                    # we don't break before the soft hyphen
-                    pango.pango_layout_set_width(layout.layout, -1)
-                    resume_index += len(soft_hyphen.encode())
                 break
 
         if not hyphenated and not first_line_text:
@@ -487,10 +468,8 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         # instance) from keeping their shape when wrapped on the next line with
         # pango layout. Maybe insert Unicode shaping characters in text?
         layout.set_text(text)
-        pango.pango_layout_set_width(
-            layout.layout, units_from_double(max_width))
-        pango.pango_layout_set_wrap(
-            layout.layout, PANGO_WRAP_MODE['WRAP_CHAR'])
+        pango.pango_layout_set_width(layout.layout, int(max_width * TO_UNITS))
+        pango.pango_layout_set_wrap(layout.layout, PANGO_WRAP_MODE['WRAP_CHAR'])
         first_line, index = layout.get_first_line()
         resume_index = index or first_line.length
         if resume_index >= len(text.encode()):
@@ -520,12 +499,18 @@ def get_log_attrs(text, lang):
     return bytestring, log_attrs
 
 
-def can_break_text(text, lang):
+def next_break_point(text, lang):
     if not text or len(text) < 2:
         return None
     bytestring, log_attrs = get_log_attrs(text, lang)
     length = len(text) + 1
-    return any(attr.is_line_break for attr in log_attrs[1:length - 1])
+    for i, attr in enumerate(log_attrs[1:length - 1]):
+        if attr.is_line_break:
+            return i
+
+
+def can_break_text(text, lang):
+    return next_break_point(text, lang) is not None
 
 
 def get_next_word_boundaries(text, lang):
